#pragma once



namespace armarx::aron
{

    /**
     * Framework for converting ARON DTOs (Data Transfer Objects) to C++ BOs
     * (Business Objects) and back.
     *
     * To allow conversion between custom ARON and C++ types, declare two
     * functions in the namespace of the BO:
     *
     * @code
     * // aron_conversions.h
     *
     * namespace bo_namespace
     * {
     *     void toAron(arondto::MyObject& dto, const MyObject& bo);
     *     void fromAron(const arondto::MyObject& dto, MyObject& bo);
     * }
     * @endcode
     *
     * Note that the DTO always comes first, and the target object is
     * non-const.
     *
     * In the implementation,
     *
     *
     * @code
     * // aron_conversions.cpp
     *
     * #include "aron_conversions.h"
     * #include <Path/to/MyValue/aron_conversions.h>
     *
     * void bo_namespace::toAron(arondto::MyObject& dto, const MyObject& bo)
     * {
     *     dto.name = bo.name;
     *     toAron(dto.myValue, bo.myValue);
     * }
     *
     * void bo_namespace::fromAron(const arondto::MyObject& dto, MyObject& bo)
     * {
     *     bo.name = dto.name;
     *     fromAron(dto.myValue, bo.myValue);
     * }
     * @endcode
     */

    // Same type
    template <class T>
    void toAron(T& dto, const T& bo)
    {
        dto = bo;
    }
    template <class T>
    void fromAron(const T& dto, T& bo)
    {
        bo = dto;
    }


    // Generic return version

    template <class DtoT, class BoT>
    DtoT toAron(const BoT& bo)
    {
        DtoT dto;
        toAron(dto, bo);
        return dto;
    }
    template <class BoT, class DtoT>
    BoT fromAron(const DtoT& dto)
    {
        BoT bo;
        fromAron(dto, bo);
        return bo;
    }

}
