#include "SkillProviderComponentPlugin.h"

#include <ArmarXCore/core/Component.h>

#include <RobotAPI/libraries/aron/core/data/variant/primitive/All.h>

namespace armarx::plugins
{
    void SkillProviderComponentPlugin::preOnInitComponent()
    {

    }

    void SkillProviderComponentPlugin::preOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        p.getProxy(myPrx, -1);
    }

    void SkillProviderComponentPlugin::postOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        skills::manager::dto::ProviderInfo i;
        i.provider = myPrx;
        i.providerName = providerName;
        i.providedSkills = p.getSkills();
        skillManager->addProvider(i);
    }

    void SkillProviderComponentPlugin::preOnDisconnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        skillManager->removeProvider(providerName);
    }

    void SkillProviderComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        std::string prefix = "skill.";
        properties->component(skillManager, "SkillMemory", prefix + "SkillManager", "The name of the SkillManager (or SkillMemory) proxy.");
    }
}


namespace armarx
{
    SkillProviderComponentPluginUser::SkillProviderComponentPluginUser()
    {
        addPlugin(plugin);
    }

    void SkillProviderComponentPluginUser::addSkill(const std::shared_ptr<skills::Skill>& skill)
    {
        if (!skill)
        {
            return;
        }

        std::lock_guard l(skillsMutex);

        std::string skillName = skill->description.skillName;
        if (skillImplementations.find(skillName) != skillImplementations.end())
        {
            ARMARX_WARNING << "Try to add a skill '" + skillName + "' which already exists in list. Ignoring this skill.";
            return;
        }

        ARMARX_INFO << "Adding skill " << skillName;
        skills::detail::SkillImplementationWrapper s(skill);
        skillImplementations.insert({skillName, s});
    }

    void SkillProviderComponentPluginUser::addSkill(const skills::helper::LambdaSkill::FunT& f, const skills::SkillDescription& desc)
    {
        auto lambda = std::make_shared<skills::helper::LambdaSkill>(f, desc);
        addSkill(lambda);
    }

    skills::provider::dto::SkillDescriptionMap SkillProviderComponentPluginUser::getSkills(const Ice::Current &current)
    {
        std::lock_guard l(skillsMutex);
        skills::provider::dto::SkillDescriptionMap skillDesciptions;
        for (const auto& [key, skillWrapper] : skillImplementations)
        {
            skillDesciptions.insert({key, skillWrapper.skill->description.toIce()});
        }
        return skillDesciptions;
    }

    skills::provider::dto::SkillStatusUpdate SkillProviderComponentPluginUser::getSkillExecutionStatus(const std::string& skill, const Ice::Current &current)
    {
        std::lock_guard l(skillsMutex);
        auto& skillWrapper = skillImplementations.at(skill);

        std::lock_guard l2(skillWrapper.skillStatusMutex);
        return skillWrapper.statusUpdate.toIce();
    }

    void SkillProviderComponentPluginUser::executeSkill(const skills::provider::dto::SkillExecutionInfo& info, const Ice::Current &current)
    {
        std::lock_guard l(skillsMutex);
        std::string skillName = info.skillName;
        ARMARX_CHECK_EXPRESSION(skillImplementations.count(skillName) > 0);

        auto& wrapper = skillImplementations.at(skillName);
        if (wrapper.task.joinable())
        {
            wrapper.task.join();
        }

        // update input params
        wrapper.reset();
        wrapper.statusUpdate.usedCallbackInterface = info.callbackInterface;
        wrapper.statusUpdate.skillName = info.skillName;
        wrapper.statusUpdate.providerName = getName();
        wrapper.statusUpdate.usedParameterization = aron::data::Dict::FromAronDictDTO(info.params);

        // recreate thread and execute skill.
        wrapper.task = std::thread{ [&] { wrapper.execute();}};
    }

    skills::provider::dto::SkillStatusUpdate SkillProviderComponentPluginUser::abortSkill(const std::string& skillName, const Ice::Current &current)
    {
        std::lock_guard l(skillsMutex);
        ARMARX_CHECK_EXPRESSION(skillImplementations.count(skillName) > 0);

        auto& wrapper = skillImplementations.at(skillName);

        std::lock_guard l2(wrapper.skillStatusMutex);

        if (wrapper.skill->running)
        {
            wrapper.skill->notifyStopped();
            wrapper.task.join();
        }

        return wrapper.statusUpdate.toIce();
    }
}
